# domain logic for interpreting the VAI json file data structures
import json
from dataclasses import dataclass
from glob import glob
from typing import Iterable, List


@dataclass
class Model:
    id: str
    name: str
    desc: str
    version: str
    files: List[str]


def parse_models_metadata(model_metas:Iterable[dict], backend:str) -> Iterable[Model]:
    def permute(blocks):
        # assumption: the blocks data structure represents inlined tuples of (height, width)
        # this means that combination 1 is (block[0], block[1]), 2 is (block[2], block[3]) and so on.
        # same scheme applies to the 'capabilities' arrays for tensor-rt
        return ((h, w) for h,w in zip(blocks[::2], blocks[1::2]))

    def derive_filename(model_id, version, net, scale, width, height, row=0, column=0):
        # e.g. fgnet-fp32-[H]x[W]-[S]x-ox.tz
        suffix = (net
                  .replace('[H]', str(height))
                  .replace('[W]', str(width))
                  .replace('[S]', str(scale))
                  .replace('[R]', str(row))
                  .replace('[C]', str(column)))
        return f'{model_id}-v{version}-{suffix}'

    def find_model_files_for_backend(model, backend):
        backend_dict = model['backends'].get(backend, {})
        capabilities = backend_dict.get('capabilities', [])
        cap_combinations = list(permute(capabilities))
        model_id = model['shortName']
        version = model['version']

        scales = backend_dict.get('scales', {})
        for scale in scales:
            for net in scales[scale].get('nets', {}):
                blocks = scales[scale].get('blocks', {})
                if blocks:
                    block_combinations = permute(blocks)
                    for width, height in block_combinations:
                        if capabilities:
                            for row, column in cap_combinations:
                                filename = derive_filename(model_id, version, net, scale, width, height,
                                                           row=row, column=column)
                                yield filename
                        else:
                            filename = derive_filename(model_id, version, net, scale, width, height)
                            yield filename

    def parse(model):
        id = model['shortName']
        desc = model.get('gui', {}).get('desc', '<no description provided by topaz>')
        files = list(find_model_files_for_backend(model, backend))
        version = model['version']

        name = model.get('gui', {}).get('name')
        if not name:
            name = model.get('displayName')
        if not name:
            name = id

        return Model(id, name, desc, version, files)

    for meta in model_metas:
        model = parse(meta)
        if model.files:
            yield model


def find_models_metadata(json_dir:str) -> dict:
    def read():
        path = json_dir.replace('\\', '/')  # glob doesn't understand windows paths

        for fn in glob(f'{path}/*.json'):
            with open(fn, 'rb') as f:
                yield json.load(f)

    for d in read():
        if 'backends' in d:
            yield d


def find_backend_files(model_dicts:Iterable[dict], backend:str, includes:str) -> Iterable[str]:
    want = set([s.strip() for s in includes.split(',') if s.strip()]) if includes else set()
    parsed = parse_models_metadata(model_dicts, backend)
    models = (model for model in parsed
              if (not want) or (f'{model.id}-{model.version}' in want))
    files = (file for model in models for file in model.files)
    return files



import os
import sys
try:
    from .cli import main
except ImportError:
    from cli import main

# give the program a descriptive name if run from python -m
if sys.argv[0].endswith('__main__.py'):
    sys.argv[0] = os.path.dirname(__file__)

main()



#!/usr/bin/env python
import argparse
import os
import platform
import sys

from textwrap import fill, indent
from typing import Iterable

from .net import download_model_files
from .vai import find_backend_files, find_models_metadata, parse_models_metadata
from .version import __version__

WINDOWS_DEFAULT_JSON_DIR = "C:\\ProgramData\\Topaz Labs LLC\\Topaz Video AI\\models"
ENV_JSON_DIR = 'TVAI_MODEL_DIR'
ENV_MODEL_DIR = 'TVAI_MODEL_DATA_DIR'
ENV_BACKEND = 'TVAI_BACKEND'
ENV_COOKIE = 'TVAI_COOKIE'
AVG_MODEL_SIZE_MB = 91

def main():
    opts = parser_args()
    metas = find_models_metadata(opts.json_dir)
    if opts.cmd == 'backends':
        list_backends(metas)
    elif opts.cmd == 'list-models':
        list_models(metas, opts.backend)
    elif opts.cmd == 'list-files':
        list_backend_files(metas, opts.backend, opts.include)
    elif opts.cmd == 'download':
        download_missing_files(metas, opts.backend, opts.include, opts.dir, opts.threads, opts.cookie)
    else:
        print(f'command {opts.cmd} not implemented (this is a bug)', file=sys.stderr)
        sys.exit(-1)


def parser_args() -> argparse.Namespace:
    win = platform.system() == 'Windows'
    json_dir = os.getenv(ENV_JSON_DIR, WINDOWS_DEFAULT_JSON_DIR if win else None)
    model_dir = os.getenv(ENV_MODEL_DIR)
    backend = os.getenv(ENV_BACKEND)
    cookie = os.getenv(ENV_COOKIE)

    helper = argparse.ArgumentDefaultsHelpFormatter

    parser = argparse.ArgumentParser(
        formatter_class=helper,
        description=f'VAI Models Manager (v{__version__}): Download missing models for TopazLabs Video AI'
    )
    parser.add_argument('--json-dir', metavar='path', default=json_dir, required=not json_dir,
                        help='Directory where the VAI model json files reside. E.g. alq-13.json. Defaults to the '
                             f'value of the environment variable {ENV_JSON_DIR} if set, else the windows default if '
                             'running on windows, else you have to specify it yourself.')

    commands = parser.add_subparsers(dest='cmd', title='commands', metavar='')
    commands.add_parser('backends', help='Lists the available backends that VAI supports')

    mlist = commands.add_parser('list-models', help='Lists available models for a given backend')
    mlist.add_argument('--backend', metavar='name', default=backend, required=not backend,
                       help=f'Name of the backend to list models for (env: {ENV_BACKEND})')

    files = commands.add_parser('list-files', help='Lists model files for a backend')
    files.add_argument('--backend', metavar='name', default=backend, required=not backend,
                       help=f'Name of the backend to list the model files for (env: {ENV_BACKEND})')
    files.add_argument('--include', metavar='ids', help='Commma separated list of specific model(s) to include')

    download = commands.add_parser('download', formatter_class=helper,
                                   help='Download VAI models missing from your model data directory')
    download.add_argument('--backend', metavar='name', default=backend, required=not backend,
                          help=f'Name of the backend to fetch models for (env: {ENV_BACKEND})')
    download.add_argument('--include', metavar='ids', help='Commma separated list of specific model(s) to include')
    download.add_argument('-d', '--dir', metavar='path', default=model_dir, required=not model_dir,
                          help='Path to your model data directory. Defaults to env-var TVAI_MODEL_DATA_DIR if set.')
    download.add_argument('-c', '--cookie', metavar='str', default=cookie, required=not cookie,
                          help='The value of the cf_clearance cookie, required to download files from topaz '
                               'Cloudflare CDN. You can find this when logged into the topaz website, by opening '
                               '"developer tools" in firefox (or inspector in chrome), then the network tab. Once '
                               'that is done, download a test model from the browser. E.g: '
                               'https://veai-models.topazlabs.com/prap-v3-fp32-ov.tz . Finally look at the request '
                               'headers for the associated request, and the Cookie header. That header has the value '
                               'required. It looks like "cf_clearance: <the-string-you-need-here>".')
    download.add_argument('-t', '--threads', metavar='n', default=4, type=int,
                          help='Number of concurrent downloads to use')

    if len(sys.argv) < 2:
        sys.argv.append('-h')
    return parser.parse_args()


def list_backends(model_dicts:Iterable[dict]):
    backends = sorted(set(b for m in model_dicts for b in m['backends']))
    print('Supported backends:', ', '.join(backends))


def list_models(model_dicts:Iterable[dict], backend:str):
    parsed = parse_models_metadata(model_dicts, backend)
    models = sorted(parsed, key=lambda m: m.id + str(m.version))

    print('Available models:')
    for m in models:
        id = f"{m.id}-{m.version}"
        desc = indent(fill(m.desc, 78), prefix=" "*11)
        print(f"\n* {id:<8s} {m.name}\n{desc}")


def list_backend_files(model_dicts:Iterable[dict], backend:str, includes:str):
    files = sorted(find_backend_files(model_dicts, backend, includes))
    print('Model files:\n', file=sys.stderr)
    print('\n'.join(files))
    print(f'\nEstimated total size: {len(files) * AVG_MODEL_SIZE_MB/(1<<10):.2f} GiB', file=sys.stderr)


def download_missing_files(model_dicts:Iterable[dict], backend:str, includes:str,
                           data_dir:str, threads:int, cookie:str):
    missing_files = find_missing_files(model_dicts, backend, includes, data_dir)
    download_model_files(missing_files, threads, cookie)


def find_missing_files(model_dicts:Iterable[dict], backend:str, includes:str, data_dir:str):
    backend_files = find_backend_files(model_dicts, backend, includes)
    target_files = (os.path.join(data_dir, file) for file in backend_files)
    missing_files = (file for file in target_files if not os.path.exists(file))
    return missing_files



import os
import sys
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import wraps
from typing import Iterable
from urllib.request import Request, urlopen, ContentTooShortError, HTTPError

from tqdm import tqdm

from .version import __version__

DOWNLOAD_URL_PREFIX = 'https://veai-models.topazlabs.com/'

def download_model_files(missing_files:Iterable[str], threads:int, cookie:str):
    def assert_download_success(future) -> int:  # bytes written
        try:
            return future.result()  # Just check that no error was generated during the download
        except BaseException as e:
            if isinstance(e, KeyboardInterrupt):
                raise
            print(f'Download failed: {e}', file=sys.stderr)
            sys.exit(1)

    def scale_unit(nbytes:int) -> str:
        if nbytes > 1<<30:
            return f'{nbytes/(1<<30):.2f} GiB'
        if nbytes > 1<<20:
            return f'{nbytes/(1<<20):.2f} MiB'
        if nbytes > 1<<10:
            return f'{nbytes/(1<<10):.2f} KiB'
        return f'{nbytes} B'

    download_jobs = ((DOWNLOAD_URL_PREFIX + os.path.basename(file), file) for file in missing_files)

    with ThreadPoolExecutor(max_workers=threads) as executor:
        futures = {executor.submit(download_file, url, cookie, file): file for url, file in download_jobs}
        if not futures:
            print('No missing model files to download')
            return

        started_at = time.time() - 0.00001
        total_bytes_downloaded = 0
        elapsed = 0

        try:
            with tqdm(total=len(futures), unit=' files ', desc='Downloading', leave=True) as pbar:
                for future in as_completed(futures):
                    total_bytes_downloaded += assert_download_success(future)
                    elapsed = time.time() - started_at
                    bps = total_bytes_downloaded / elapsed
                    speed = f'{scale_unit(bps)}/s'
                    pbar.update(1)
                    pbar.set_postfix({'speed': speed})

        except KeyboardInterrupt:
            for future in futures:
                future.cancel()
            executor.shutdown(wait=False)
            raise

    print(f'Download completed successfully with {scale_unit(total_bytes_downloaded)} of data fetched.')


def retry(max_tries=5, initial_wait_secs=1, backoff=3):
    def decorator(func):
        @wraps(func)
        def inner(*args, **kw):
            wait_duration = initial_wait_secs
            for attempt in range(max_tries):
                try:
                    return func(*args, **kw)
                except (HTTPError, ContentTooShortError) as exc:
                    if isinstance(exc, HTTPError) and 400 <= exc.code < 500 and exc.code not in (408,429):
                        raise  # don't retry on file not found, or invalid cf-cookie
                    if attempt == max_tries - 1:
                        raise
                    print(f'Error encountered during download: {exc}, Retrying in {wait_duration} seconds...',
                          file=sys.stderr)
                    time.sleep(wait_duration)
                    wait_duration *= backoff
        return inner
    return decorator


@retry(max_tries=5)
def download_file(url:str, cookie:str, filename:str, chunk_size:int=8<<10) -> int:
    request = Request(url, headers={
        'cookie': f'cf_clearance={cookie}',
        'accept': '*/*',
        'user-agent': f'vaimm/{__version__}',  # for some reason cloudflare rejects python's UA !?
        'accept-encoding': '*',
    })

    tempfile = f'{filename}.incomplete'
    bytes_written = 0

    with urlopen(request) as response:
        expected_bytes = response.headers.get('content-length')
        with open(tempfile, 'wb') as f:
            while True:
                chunk = response.read(chunk_size)
                if not chunk:  # end of data in stream
                    break
                f.write(chunk)
                bytes_written += len(chunk)

    if expected_bytes is not None:
        expected_bytes = int(expected_bytes)
        if expected_bytes != bytes_written:
            raise ContentTooShortError(
                f'Incomplete data from server - expected: {expected_bytes}, got: {bytes_written} for url: {url}', None)

    # ensure user doesn't end up with broken models in case something goes wrong / program is aborted mid-stride
    os.rename(tempfile, filename)
    return bytes_written
